@@ -2,21 +2,21 @@
  *
  * logging.c - message logging
  *
- * Copyright © 2007 Scott James Remnant <scott@netsplit.com>.
+ * Copyright © 2009 Scott James Remnant <scott@netsplit.com>.
+ * Copyright © 2009 Canonical Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * it under the terms of the GNU General Public License version 2, as
+ * published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -30,6 +30,7 @@
 #include <string.h>
 #include <syslog.h>
 
+#include <nih/macros.h>
 #include <nih/alloc.h>
 #include <nih/string.h>
 #include <nih/main.h>
@@ -58,7 +59,7 @@ NihLogLevel nih_log_priority = NIH_LOG_UNKNOWN;
  *
  * Initialise the default logger and priority.
  **/
-static inline void
+void
 nih_log_init (void)
 {
 	if (! logger)
@@ -117,13 +118,13 @@ nih_log_set_priority (NihLogLevel new_priority)
  * to being below the minimum priority and negative value if the logger failed.
  **/
 int
-nih_log_message (NihLogLevel  priority,
-		 const char  *format,
+nih_log_message (NihLogLevel priority,
+		 const char *format,
 		 ...)
 {
-	char    *message = NULL;
-	va_list  args;
-	int      ret;
+	nih_local char *message = NULL;
+	va_list         args;
+	int             ret;
 
 	nih_assert (format != NULL);
 
@@ -133,12 +134,11 @@ nih_log_message (NihLogLevel  priority,
 		return 1;
 
 	va_start (args, format);
-	NIH_MUST (message = nih_vsprintf (NULL, format, args));
+	message = NIH_MUST (nih_vsprintf (NULL, format, args));
 	va_end (args);
 
 	/* Output the message */
 	ret = logger (priority, message);
-	nih_free (message);
 
 	return ret;
 }
@@ -155,8 +155,8 @@ nih_log_message (NihLogLevel  priority,
  * Returns: zero on completion, negative value on error.
  **/
 int
-nih_logger_printf (NihLogLevel  priority,
-		   const char  *message)
+nih_logger_printf (NihLogLevel priority,
+		   const char *message)
 {
 	nih_assert (message != NULL);
 
@@ -200,8 +200,8 @@ nih_logger_printf (NihLogLevel  priority,
  * Returns: zero on completion, negative value on error.
  **/
 int
-nih_logger_syslog (NihLogLevel  priority,
-		   const char  *message)
+nih_logger_syslog (NihLogLevel priority,
+		   const char *message)
 {
 	int level;
 
