@@ -2,21 +2,21 @@
  *
  * test_logging.c - test suite for nih/logging.c
  *
- * Copyright © 2007 Scott James Remnant <scott@netsplit.com>.
+ * Copyright © 2009 Scott James Remnant <scott@netsplit.com>.
+ * Copyright © 2009 Canonical Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * it under the terms of the GNU General Public License version 2, as
+ * published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include <nih/test.h>
@@ -32,14 +32,16 @@
 
 
 static NihLogLevel last_priority = NIH_LOG_UNKNOWN;
-static char *last_message = NULL;
+static char *      last_message = NULL;
 
 static int
-my_logger (NihLogLevel  priority,
-	   const char  *message)
+my_logger (NihLogLevel priority,
+	   const char *message)
 {
-	last_priority = priority;
-	last_message = strdup (message);
+	TEST_ALLOC_SAFE {
+		last_priority = priority;
+		last_message = strdup (message);
+	}
 
 	if (! strcmp (message, "this should error"))
 		return -1;
@@ -54,18 +56,20 @@ test_set_logger (void)
 	 * function is called for the logging macros that follow.
 	 */
 	TEST_FUNCTION ("nih_log_set_logger");
-	nih_log_set_priority (NIH_LOG_MESSAGE);
-	nih_log_set_logger (my_logger);
+	TEST_ALLOC_FAIL {
+		nih_log_set_priority (NIH_LOG_MESSAGE);
+		nih_log_set_logger (my_logger);
 
-	last_priority = NIH_LOG_UNKNOWN;
+		last_priority = NIH_LOG_UNKNOWN;
 
-	nih_fatal ("some message");
+		nih_fatal ("some message");
 
-	TEST_EQ (last_priority, NIH_LOG_FATAL);
+		TEST_EQ (last_priority, NIH_LOG_FATAL);
 
-	free (last_message);
+		free (last_message);
 
-	nih_log_set_logger (nih_logger_printf);
+		nih_log_set_logger (nih_logger_printf);
+	}
 }
 
 void
@@ -75,19 +79,21 @@ test_set_priority (void)
 	 * log messages we wouldn't have previously received.
 	 */
 	TEST_FUNCTION ("nih_log_set_priority");
-	nih_log_set_logger (my_logger);
-	nih_log_set_priority (NIH_LOG_DEBUG);
+	TEST_ALLOC_FAIL {
+		nih_log_set_logger (my_logger);
+		nih_log_set_priority (NIH_LOG_DEBUG);
 
-	last_priority = NIH_LOG_UNKNOWN;
+		last_priority = NIH_LOG_UNKNOWN;
 
-	nih_debug ("some message");
+		nih_debug ("some message");
 
-	TEST_EQ (last_priority, NIH_LOG_DEBUG);
+		TEST_EQ (last_priority, NIH_LOG_DEBUG);
 
-	free (last_message);
+		free (last_message);
 
-	nih_log_set_logger (nih_logger_printf);
-	nih_log_set_priority (NIH_LOG_MESSAGE);
+		nih_log_set_logger (nih_logger_printf);
+		nih_log_set_priority (NIH_LOG_MESSAGE);
+	}
 }
 
 void
@@ -138,11 +144,13 @@ test_log_message (void)
 	 * by the function.
 	 */
 	TEST_FEATURE ("with error code returned from logger");
-	ret = nih_log_message (NIH_LOG_FATAL, "this should error");
+	TEST_ALLOC_FAIL {
+		ret = nih_log_message (NIH_LOG_FATAL, "this should error");
 
-	TEST_LT (ret, 0);
+		TEST_LT (ret, 0);
 
-	free (last_message);
+		free (last_message);
+	}
 
 
 	/* Check that the nih_debug macro wraps the call properly and
@@ -259,7 +267,6 @@ test_logger_printf (void)
 
 	TEST_FUNCTION ("nih_logger_printf");
 	output = tmpfile ();
-	program_name = "test";
 	nih_log_set_priority (NIH_LOG_DEBUG);
 
 
@@ -267,104 +274,116 @@ test_logger_printf (void)
 	 * correctly, without the program name prefixed to the front.
 	 */
 	TEST_FEATURE ("with low priority message");
-	TEST_DIVERT_STDOUT (output) {
-		ret = nih_log_message (NIH_LOG_DEBUG,
-				       "message with %s %d formatting",
-				       "some", 20);
+	TEST_ALLOC_FAIL {
+		TEST_DIVERT_STDOUT (output) {
+			ret = nih_log_message (NIH_LOG_DEBUG,
+					       "message with %s %d formatting",
+					       "some", 20);
+		}
+		rewind (output);
+
+		TEST_EQ (ret, 0);
+		TEST_FILE_EQ (output, "message with some 20 formatting\n");
+		TEST_FILE_END (output);
+
+		TEST_FILE_RESET (output);
 	}
-	rewind (output);
-
-	TEST_EQ (ret, 0);
-	TEST_FILE_EQ (output, "message with some 20 formatting\n");
-	TEST_FILE_END (output);
-
-	TEST_FILE_RESET (output);
 
 
 	/* Check that ordinary messages are output on stdout and formatted
 	 * correctly, without the program name prefixed on the front.
 	 */
 	TEST_FEATURE ("with ordinary message");
-	TEST_DIVERT_STDOUT (output) {
-		ret = nih_log_message (NIH_LOG_MESSAGE,
-				       "message with %s %d formatting",
-				       "some", 20);
+	TEST_ALLOC_FAIL {
+		TEST_DIVERT_STDOUT (output) {
+			ret = nih_log_message (NIH_LOG_MESSAGE,
+					       "message with %s %d formatting",
+					       "some", 20);
+		}
+		rewind (output);
+
+		TEST_EQ (ret, 0);
+		TEST_FILE_EQ (output, "message with some 20 formatting\n");
+		TEST_FILE_END (output);
+
+		TEST_FILE_RESET (output);
 	}
-	rewind (output);
-
-	TEST_EQ (ret, 0);
-	TEST_FILE_EQ (output, "message with some 20 formatting\n");
-	TEST_FILE_END (output);
-
-	TEST_FILE_RESET (output);
 
 
 	/* Check that warning messages are output on stderr and formatted
 	 * correctly, with the program name prefixed on the front.
 	 */
 	TEST_FEATURE ("with ordinary message");
-	TEST_DIVERT_STDERR (output) {
-		ret = nih_log_message (NIH_LOG_WARN,
-				       "message with %s %d formatting",
-				       "some", 20);
+	TEST_ALLOC_FAIL {
+		TEST_DIVERT_STDERR (output) {
+			ret = nih_log_message (NIH_LOG_WARN,
+					       "message with %s %d formatting",
+					       "some", 20);
+		}
+		rewind (output);
+
+		TEST_EQ (ret, 0);
+		TEST_FILE_EQ (output, "test: message with some 20 formatting\n");
+		TEST_FILE_END (output);
+
+		TEST_FILE_RESET (output);
 	}
-	rewind (output);
-
-	TEST_EQ (ret, 0);
-	TEST_FILE_EQ (output, "test: message with some 20 formatting\n");
-	TEST_FILE_END (output);
-
-	TEST_FILE_RESET (output);
 
 
 	/* Check that a high priority message is output to stderr and formatted
 	 * correctly, with the program name prefixed on the front.
 	 */
 	TEST_FEATURE ("with high priority message");
-	TEST_DIVERT_STDERR (output) {
-		ret = nih_log_message (NIH_LOG_FATAL,
-				       "%s message %d formatted",
-				       "error", -1);
+	TEST_ALLOC_FAIL {
+		TEST_DIVERT_STDERR (output) {
+			ret = nih_log_message (NIH_LOG_FATAL,
+					       "%s message %d formatted",
+					       "error", -1);
+		}
+		rewind (output);
+
+		TEST_EQ (ret, 0);
+		TEST_FILE_EQ (output, "test: error message -1 formatted\n");
+		TEST_FILE_END (output);
+
+		TEST_FILE_RESET (output);
 	}
-	rewind (output);
-
-	TEST_EQ (ret, 0);
-	TEST_FILE_EQ (output, "test: error message -1 formatted\n");
-	TEST_FILE_END (output);
-
-	TEST_FILE_RESET (output);
 
 
 	/* Check that a message with a GNU-style filename prefix is correctly
 	 * formatted, with no space between the program name and the message.
 	 */
 	TEST_FEATURE ("with prefixed message");
-	TEST_DIVERT_STDERR (output) {
-		ret = nih_log_message (NIH_LOG_FATAL,
-				       "%s:%d: some error or other",
-				       "example.txt", 303);
+	TEST_ALLOC_FAIL {
+		TEST_DIVERT_STDERR (output) {
+			ret = nih_log_message (NIH_LOG_FATAL,
+					       "%s:%d: some error or other",
+					       "example.txt", 303);
+		}
+		rewind (output);
+
+		TEST_EQ (ret, 0);
+		TEST_FILE_EQ (output, "test:example.txt:303: some error or other\n");
+		TEST_FILE_END (output);
+
+		TEST_FILE_RESET (output);
 	}
-	rewind (output);
-
-	TEST_EQ (ret, 0);
-	TEST_FILE_EQ (output, "test:example.txt:303: some error or other\n");
-	TEST_FILE_END (output);
-
-	TEST_FILE_RESET (output);
 
 
 	/* Check that an error code is returned if the output stream is
 	 * closed.
 	 */
 	TEST_FEATURE ("with closed stream");
-	TEST_DIVERT_STDERR (output) {
-		close (STDERR_FILENO);
+	TEST_ALLOC_FAIL {
+		TEST_DIVERT_STDERR (output) {
+			close (STDERR_FILENO);
 
-		ret = nih_log_message (NIH_LOG_FATAL, "an error message");
+			ret = nih_log_message (NIH_LOG_FATAL, "an error message");
+		}
+
+		TEST_LT (ret, 0);
 	}
 
-	TEST_LT (ret, 0);
-
 
 	fclose (output);
 	nih_log_set_priority (NIH_LOG_MESSAGE);
@@ -375,6 +394,8 @@ int
 main (int   argc,
       char *argv[])
 {
+	program_name = "test";
+
 	test_set_logger ();
 	test_set_priority ();
 	test_log_message ();
