@@ -1,21 +1,20 @@
 /* upstart
  *
- * Copyright © 2006 Canonical Ltd.
- * Author: Scott James Remnant <scott@ubuntu.com>.
+ * Copyright © 2009 Canonical Ltd.
+ * Author: Scott James Remnant <scott@netsplit.com>.
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * it under the terms of the GNU General Public License version 2, as
+ * published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -23,13 +22,15 @@
 #endif /* HAVE_CONFIG_H */
 
 
+#include <dbus/dbus.h>
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/param.h>
 
 #include <pwd.h>
-#include <utmp.h>
 #include <time.h>
+#include <utmpx.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <stdio.h>
@@ -44,21 +45,17 @@
 #include <nih/timer.h>
 #include <nih/string.h>
 #include <nih/signal.h>
+#include <nih/io.h>
 #include <nih/main.h>
 #include <nih/option.h>
 #include <nih/logging.h>
 #include <nih/error.h>
 
-#include <upstart/control.h>
+#include <nih-dbus/dbus_error.h>
+#include <nih-dbus/errors.h>
 
+#include "sysv.h"
 
-/**
- * PID_FILE:
- *
- * Where we store our process id so that a running shutdown can be detected
- * or cancelled.
- **/
-#define PID_FILE "/var/run/shutdown.pid"
 
 /**
  * ETC_NOLOGIN:
@@ -84,8 +81,10 @@
 #endif
 
 
+/* Prototypes for option functions */
+static int runlevel_option (NihOption *option, const char *arg);
+
 /* Prototypes for static functions */
-static int   event_setter      (NihOption *option, const char *arg);
 static void  shutdown_now      (void)
 	__attribute__ ((noreturn));
 static void  cancel_callback   (void *data, NihSignal *signal)
@@ -98,23 +97,30 @@ static void  sysvinit_shutdown (void);
 
 
 /**
- * event:
+ * runlevel:
  *
- * Event to send after the shutdown event.
+ * Runlevel to switch to.
  **/
-static char *event = NULL;
+static int runlevel = 0;
+
+/**
+ * init_halt:
+ *
+ * Value of init_halt environment variable for event.
+ **/
+static const char *init_halt = NULL;
 
 /**
  * cancel:
  *
- * True if we should cancel an already running shutdown.
+ * TRUE if we should cancel an already running shutdown.
  **/
 static int cancel = FALSE;
 
 /**
  * warn_only:
  *
- * True if we should only send the warning, and not perform the actual
+ * TRUE if we should only send the warning, and not perform the actual
  * shutdown.
  **/
 static int warn_only = FALSE;
@@ -134,25 +140,67 @@ static char *when = NULL;
 static int delay = 0;
 
 
+/**
+ * runlevel_option:
+ * @option: option found in arguments,
+ * @arg: always NULL.
+ *
+ * This function is called whenever one of the -r, -h, -H or -P options
+ * is found in the argument list.  It changes the runlevel to that implied
+ * by the option.
+ **/
+static int
+runlevel_option (NihOption  *option,
+		 const char *arg)
+{
+	int *value;
+
+	nih_assert (option != NULL);
+	nih_assert (option->value != NULL);
+	nih_assert (arg == NULL);
+
+	value = (int *)option->value;
+
+	switch (option->option) {
+	case 'r':
+		*value = '6';
+		init_halt = NULL;
+		break;
+	case 'h':
+		*value = '0';
+		init_halt = NULL;
+		break;
+	case 'H':
+		*value = '0';
+		init_halt = "HALT";
+		break;
+	case 'P':
+		*value = '0';
+		init_halt = "POWEROFF";
+		break;
+	}
+
+	return 0;
+}
+
+
 /**
  * options:
  *
  * Command-line options accepted for all arguments.
  **/
 static NihOption options[] = {
-	{ 'e', "event", N_("shutdown event to send"),
-	  NULL, "EVENT", &event, NULL },
-	{ 'r', "reboot", N_("reboot after shutdown"),
-	  NULL, NULL, &event, event_setter },
-	{ 'h', "shutdown", N_("halt or power off after shutdown"),
-	  NULL, NULL, &event, event_setter },
-	{ 'H', "halt", N_("halt after shutdown (implies -h)"),
-	  NULL, NULL, &event, event_setter },
-	{ 'P', "poweroff", N_("power off after shutdown (implies -h)"),
-	  NULL, NULL, &event, event_setter },
-	{ 'c', "cancel", N_("cancel a running shutdown"),
+	{ 'r', NULL, N_("reboot after shutdown"),
+	  NULL, NULL, &runlevel, runlevel_option },
+	{ 'h', NULL, N_("halt or power off after shutdown"),
+	  NULL, NULL, &runlevel, runlevel_option },
+	{ 'H', NULL, N_("halt after shutdown (implies -h)"),
+	  NULL, NULL, &runlevel, runlevel_option },
+	{ 'P', NULL, N_("power off after shutdown (implies -h)"),
+	  NULL, NULL, &runlevel, runlevel_option },
+	{ 'c', NULL, N_("cancel a running shutdown"),
 	  NULL, NULL, &cancel, NULL },
-	{ 'k', "warn-only", N_("only send warnings, don't shutdown"),
+	{ 'k', NULL, N_("only send warnings, don't shutdown"),
 	  NULL, NULL, &warn_only, NULL },
 
 	/* Compatibility option for specifying time */
@@ -175,22 +223,52 @@ int
 main (int   argc,
       char *argv[])
 {
-	char   **args, *message, *msg;
-	int      arg;
-	size_t   messagelen;
-	FILE    *pidfile;
-	pid_t    pid = 0;
+	char **         args;
+	nih_local char *message = NULL;
+	size_t          messagelen;
+	nih_local char *msg = NULL;
+	int             arg;
+	pid_t           pid = 0;
 
 	nih_main_init (argv[0]);
-	nih_option_set_usage ("TIME [MESSAGE]");
+
+	nih_option_set_usage (_("TIME [MESSAGE]"));
+	nih_option_set_synopsis (_("Bring the system down."));
+	nih_option_set_help (
+		_("TIME may have different formats, the most common is simply "
+		  "the word 'now' which will bring the system down "
+		  "immediately.  Other valid formats are +m, where m is the "
+		  "number of minutes to wait until shutting down and hh:mm "
+		  "which specifies the time on the 24hr clock.\n"
+		  "\n"
+		  "Logged in users are warned by a message sent to their "
+		  "terminal, you may include an optional MESSAGE included "
+		  "with this.  Messages can be sent without actually "
+		  "bringing the system down by using the -k option.\n"
+		  "\n"
+		  "If TIME is given, the command will remain in the "
+		  "foreground until the shutdown occurs.  It can be cancelled "
+		  "by Control-C, or by another user using the -c option.\n"
+		  "\n"
+		  "The system is brought down into maintenance (single-user) "
+		  "mode by default, you can change this with either the -r or "
+		  "-h option which specify a reboot or system halt "
+		  "respectively.  The -h option can be further modified with "
+		  "-H or -P to specify whether to halt the system, or to "
+		  "power it off afterwards.  The default is left up to the "
+		  "shutdown scripts."));
 
 	args = nih_option_parser (NULL, argc, argv, options, FALSE);
 	if (! args)
 		exit (1);
 
-	/* If the event wasn't given explicitly, set it to maintenance */
-	if (! event)
-		NIH_MUST (event = nih_strdup (NULL, "maintenance"));
+	/* If the runlevel wasn't given explicitly, set it to 1 so we go
+	 * down into single-user mode.
+	 */
+	if (! runlevel) {
+		runlevel = '1';
+		init_halt = NULL;
+	}
 
 
 	/* When may be specified with -g, or must be first argument */
@@ -199,7 +277,7 @@ main (int   argc,
 		nih_main_suggest_help ();
 		exit (1);
 	} else if (! (cancel || when)) {
-		NIH_MUST (when = nih_strdup (NULL, args[0]));
+		when = NIH_MUST (nih_strdup (NULL, args[0]));
 		arg = 1;
 	} else {
 		arg = 0;
@@ -262,15 +340,12 @@ main (int   argc,
 	 * Really this should be just the next argument, but that's not
 	 * how this has been traditionally done *sigh*
 	 */
-	NIH_MUST (message = nih_strdup (NULL, ""));
+	message = NIH_MUST (nih_strdup (NULL, ""));
 	messagelen = 0;
 	for (; args[arg]; arg++) {
-		char *new_message;
-
-		NIH_MUST (new_message = nih_realloc (
+		message = NIH_MUST (nih_realloc (
 				  message, NULL,
 				  messagelen + strlen(args[arg]) + 4));
-		message = new_message;
 
 		strcat (message, args[arg]);
 		strcat (message, " ");
@@ -285,18 +360,14 @@ main (int   argc,
 	/* Check we're root, or setuid root */
 	setuid (geteuid ());
 	if (getuid ()) {
-		nih_error (_("Need to be root"));
+		nih_fatal (_("Need to be root"));
 		exit (1);
 	}
 
-	/* Look for an existing pid file */
-	pidfile = fopen (PID_FILE, "r");
-	if (pidfile) {
-		fscanf (pidfile, "%d", &pid);
-		fclose (pidfile);
-	}
-
-	/* Deal with the existing process */
+	/* Look for an existing pid file and deal with the existing
+	 * process if there is one.
+	 */
+	pid = nih_main_read_pidfile ();
 	if (pid > 0) {
 		if (cancel) {
 			if (kill (pid, SIGINT) < 0) {
@@ -318,16 +389,17 @@ main (int   argc,
 	}
 
 	/* Send an initial message */
-	NIH_MUST (msg = warning_message (message));
+	msg = NIH_MUST (warning_message (message));
 	wall (msg);
-	nih_free (msg);
 
 	if (warn_only)
 		exit (0);
 
 
 	/* Give us a sane environment */
-	chdir ("/");
+	if (chdir ("/") < 0)
+		nih_warn ("%s: %s", _("Unable to change directory"),
+			  strerror (errno));
 	umask (022);
 
 	/* Shutdown now? */
@@ -335,13 +407,13 @@ main (int   argc,
 		shutdown_now ();
 
 	/* Save our pid so we can be interrupted later */
-	unlink (PID_FILE);
-	pidfile = fopen (PID_FILE, "w");
-	if (pidfile) {
-		fprintf (pidfile, "%d\n", getpid ());
-		fclose (pidfile);
-	} else {
-		nih_warn (_("Unable to save pid to %s"), PID_FILE);
+	if (nih_main_write_pidfile (getpid ()) < 0) {
+		NihError *err;
+
+		err = nih_error_get ();
+		nih_warn ("%s: %s: %s", nih_main_get_pidfile(),
+			  _("Unable to write pid file"), err->message);
+		nih_free (err);
 	}
 
 
@@ -354,15 +426,19 @@ main (int   argc,
 
 	/* Catch the usual quit signals */
 	nih_signal_set_handler (SIGINT, nih_signal_handler);
-	nih_signal_add_callback (NULL, SIGINT, cancel_callback, NULL);
+	NIH_MUST (nih_signal_add_handler (NULL, SIGINT,
+					  cancel_callback, NULL));
 	nih_signal_set_handler (SIGQUIT, nih_signal_handler);
-	nih_signal_add_callback (NULL, SIGQUIT, cancel_callback, NULL);
+	NIH_MUST (nih_signal_add_handler (NULL, SIGQUIT,
+					  cancel_callback, NULL));
 	nih_signal_set_handler (SIGTERM, nih_signal_handler);
-	nih_signal_add_callback (NULL, SIGTERM, cancel_callback, NULL);
+	NIH_MUST (nih_signal_add_handler (NULL, SIGTERM,
+					  cancel_callback, NULL));
 
 	/* Call a timer every minute until we shutdown */
-	nih_timer_add_periodic (NULL, 60,
-				(NihTimerCb)timer_callback, message);
+	NIH_MUST (nih_timer_add_periodic (NULL, 60,
+					  (NihTimerCb)timer_callback,
+					  message));
 
 	/* Hang around */
 	nih_main_loop ();
@@ -371,49 +447,6 @@ main (int   argc,
 }
 
 
-/**
- * event_setter:
- * @option: option found in arguments,
- * @arg: always %NULL.
- *
- * This function is called whenever one of the -r, -h, -H or -P options
- * is found in the argument list.  It changes the event (which can also
- * be set by -e/--event) to that implied by the option.
- **/
-static int
-event_setter (NihOption  *option,
-	      const char *arg)
-{
-	char **value;
-
-	nih_assert (option != NULL);
-	nih_assert (option->value != NULL);
-	nih_assert (arg == NULL);
-
-	value = (char **)option->value;
-
-	if (*value)
-		nih_free (*value);
-
-	switch (option->option) {
-	case 'r':
-		NIH_MUST (*value = nih_strdup (NULL, "reboot"));
-		break;
-	case 'h':
-		NIH_MUST (*value = nih_strdup (NULL, "halt"));
-		break;
-	case 'H':
-		NIH_MUST (*value = nih_strdup (NULL, "system-halt"));
-		break;
-	case 'P':
-		NIH_MUST (*value = nih_strdup (NULL, "power-off"));
-		break;
-	}
-
-	return 0;
-}
-
-
 /**
  * shutdown_now:
  *
@@ -424,43 +457,38 @@ event_setter (NihOption  *option,
 static void
 shutdown_now (void)
 {
-	UpstartMsg msg;
-	int        sock;
+	nih_local char **extra_env = NULL;
+	NihDBusError *   dbus_err;
 
-	/* Connect to the daemon */
-	sock = upstart_open ();
-	if (sock < 0) {
-		NihError *err;
+	if (init_halt) {
+		char *e;
 
-		err = nih_error_get ();
-		nih_error (_("Unable to establish control socket: %s"),
-			   err->message);
-		exit (1);
+		e = NIH_MUST (nih_sprintf (NULL, "INIT_HALT=%s", init_halt));
+
+		extra_env = NIH_MUST (nih_str_array_new (NULL));
+		NIH_MUST (nih_str_array_addp (&extra_env, NULL, NULL, e));
 	}
 
-	/* Build the message to send */
-	msg.type = UPSTART_SHUTDOWN;
-	msg.shutdown.name = event;
+	if (sysv_change_runlevel (runlevel, extra_env, NULL, NULL) < 0) {
+		dbus_err = (NihDBusError *)nih_error_get ();
 
-	/* Send the message */
-	if (upstart_send_msg (sock, &msg) < 0) {
-		NihError *err;
+		if ((dbus_err->number != NIH_DBUS_ERROR)
+		    || strcmp (dbus_err->name, DBUS_ERROR_NO_SERVER)) {
+			nih_fatal ("%s", dbus_err->message);
+			exit (1);
+		}
 
-		err = nih_error_get ();
+		nih_free (dbus_err);
 
 		/* Connection Refused means that init isn't running, this
 		 * might mean we've just upgraded to upstart and haven't
 		 * yet rebooted ... so try /dev/initctl
 		 */
-		if (err->number == ECONNREFUSED)
-			sysvinit_shutdown ();
-
-		nih_error (_("Unable to send message: %s"), err->message);
-		exit (1);
+		sysvinit_shutdown ();
 	}
 
-	unlink (PID_FILE);
 	unlink (ETC_NOLOGIN);
+	nih_main_unlink_pidfile ();
 
 	exit (0);
 }
@@ -480,8 +508,8 @@ cancel_callback (void      *data,
 		 NihSignal *signal)
 {
 	nih_error (_("Shutdown cancelled"));
-	unlink (PID_FILE);
 	unlink (ETC_NOLOGIN);
+	nih_main_unlink_pidfile ();
 	exit (0);
 }
 
@@ -498,11 +526,11 @@ cancel_callback (void      *data,
 static void
 timer_callback (const char *message)
 {
-	char *msg;
-	int   warn = FALSE;
+	nih_local char *msg = NULL;
+	int             warn = FALSE;
 
 	delay--;
-	NIH_MUST (msg = warning_message (message));
+	msg = NIH_MUST (warning_message (message));
 
 
 	/* Write /etc/nologin with less than 5 minutes remaining */
@@ -533,8 +561,6 @@ timer_callback (const char *message)
 	/* Shutdown the machine at zero */
 	if (! delay)
 		shutdown_now ();
-
-	nih_free (msg);
 }
 
 
@@ -550,29 +576,70 @@ timer_callback (const char *message)
 static char *
 warning_message (const char *message)
 {
-	char *banner, *msg;
+	nih_local char *banner = NULL;
+	char *          msg;
 
 	nih_assert (message != NULL);
 
-	if (delay > 1) {
-		banner = nih_sprintf (NULL, _("The system is going down for "
-					      "%s in %d minutes!"),
-				      event, delay);
-	} else if (delay) {
-		banner = nih_sprintf (NULL, _("The system is going down for "
-					      "%s IN ONE MINUTE!"),
-				      event);
-	} else {
-		banner = nih_sprintf (NULL, _("The system is going down for "
-					      "%s NOW!"),
-				      event);
+	if ((runlevel == '0')
+	    && init_halt && (! strcmp (init_halt, "POWEROFF"))) {
+		if (delay) {
+			banner = nih_sprintf (
+				NULL, _n("The system is going down for "
+					 "power off in %d minute!",
+					 "The system is going down for "
+					 "power off in %d minutes!",
+					 delay), delay);
+		} else {
+			banner = nih_strdup (
+				NULL, _("The system is going down for "
+					"power off NOW!"));
+		}
+	} else if (runlevel == '0') {
+		if (delay) {
+			banner = nih_sprintf (
+				NULL, _n("The system is going down for "
+					 "halt in %d minute!",
+					 "The system is going down for "
+					 "halt in %d minutes!",
+					 delay), delay);
+		} else {
+			banner = nih_strdup (
+				NULL, _("The system is going down for "
+					"halt NOW!"));
+		}
+	} else if (runlevel == '1') {
+		if (delay) {
+			banner = nih_sprintf (
+				NULL, _n("The system is going down for "
+					 "maintenance in %d minute!",
+					 "The system is going down for "
+					 "maintenance in %d minutes!",
+					 delay), delay);
+		} else {
+			banner = nih_strdup (
+				NULL, _("The system is going down for "
+					"maintenance NOW!"));
+		}
+	} else if (runlevel == '6') {
+		if (delay) {
+			banner = nih_sprintf (
+				NULL, _n("The system is going down for "
+					 "reboot in %d minute!",
+					 "The system is going down for "
+					 "reboot in %d minutes!",
+					 delay), delay);
+		} else {
+			banner = nih_strdup (
+				NULL, _("The system is going down for "
+					"reboot NOW!"));
+		}
 	}
 
 	if (! banner)
 		return NULL;
 
 	msg = nih_sprintf (NULL, "\r%s\r\n%s", banner, message);
-	nih_free (banner);
 
 	return msg;
 }
@@ -598,12 +665,16 @@ alarm_handler (int signum)
 static void
 wall (const char *message)
 {
-	struct sigaction  act;
-	struct utmp      *ent;
-	pid_t             pid;
-	time_t            now;
-	struct tm        *tm;
-	char             *user, *tty, hostname[MAXHOSTNAMELEN], *banner;
+	struct sigaction act;
+	struct utmpx *   ent;
+	pid_t            pid;
+	time_t           now;
+	struct tm *      tm;
+	char *           user;
+	char *           tty;
+	char             hostname[MAXHOSTNAMELEN];
+	char *           banner1;
+	char *           banner2;
 
 	pid = fork ();
 	if (pid < 0) {
@@ -632,7 +703,7 @@ wall (const char *message)
 	}
 	if (! user) {
 		if (getuid ()) {
-			NIH_MUST (user = nih_sprintf (NULL, "uid %d",
+			user = NIH_MUST (nih_sprintf (NULL, "uid %d",
 						      getuid ()));
 		} else {
 			user = "root";
@@ -652,15 +723,15 @@ wall (const char *message)
 	tm = localtime (&now);
 
 	/* Construct banner */
-	NIH_MUST (banner = nih_sprintf (
-			  NULL, _("\007\r\nBroadcast message from %s@%s\r\n"
-				  "\t(%s) at %d:%02d ...\r\n\r\n"),
-			  user, hostname, tty, tm->tm_hour, tm->tm_min));
+	banner1 = nih_sprintf (NULL, _("Broadcast message from %s@%s"),
+			       user, hostname);
+	banner2 = nih_sprintf (NULL, _("(%s) at %d:%02d ..."),
+			       tty, tm->tm_hour, tm->tm_min);
 
 
 	/* Iterate entries in the utmp file */
-	setutent ();
-	while ((ent = getutent ()) != NULL) {
+	setutxent ();
+	while ((ent = getutxent ()) != NULL) {
 		char dev[PATH_MAX + 1];
 		int  fd;
 
@@ -684,7 +755,8 @@ wall (const char *message)
 
 			term = fdopen (fd, "w");
 			if (term) {
-				fputs (banner, term);
+				fprintf (term, "\007\r\n%s\r\n\t%s\r\n\r\n",
+					 banner1, banner2);
 				fputs (message, term);
 				fflush (term);
 				fclose (term);
@@ -692,7 +764,10 @@ wall (const char *message)
 		}
 		alarm (0);
 	}
-	endutent ();
+	endutxent ();
+
+	nih_free (banner1);
+	nih_free (banner2);
 
 	exit (0);
 }
@@ -731,13 +806,7 @@ sysvinit_shutdown (void)
 	request.cmd = 1;
 
 	/* Select a runlevel based on the event name */
-	if (! strcmp (event, "maintenance")) {
-		request.runlevel = '1';
-	} else if (! strcmp (event, "reboot")) {
-		request.runlevel = '6';
-	} else {
-		request.runlevel = '0';
-	}
+	request.runlevel = runlevel;
 
 
 	/* Break syscalls with SIGALRM */
